# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List

import numpy as np
from sklearn.cluster import KMeans

import model_compression_toolkit.common.quantization.quantization_config as qc
from model_compression_toolkit.common.constants import CLUSTER_CENTERS, MIN_THRESHOLD, SCALE_PER_CHANNEL, \
    MULTIPLIER_N_BITS
from model_compression_toolkit.common.quantization.quantizers.quantizers_helpers import kmeans_assign_clusters, \
    int_quantization_with_scale


def lut_kmeans_tensor(tensor_data: np.ndarray,
                      p: int,
                      n_bits: int,
                      per_channel: bool = False,
                      channel_axis: int = 1,
                      n_iter: int = 10,
                      min_threshold: float = MIN_THRESHOLD,
                      quant_error_method: qc.QuantizationErrorMethod = None) -> dict:
    """
    The quantizer first finds the closest power-of-two number to the max value per channel of tensor_data.
    Now, we divide tensor_data with the scale vector per channel. In addition, we scale the result to the range
    [-2^(MULTIPLIER_N_BITS-1), 2^(MULTIPLIER_N_BITS-1)-1].
    Next, we take the scaled tensor_data and perform k-means clustering with 2^nbit clusters.
    We return the rounded cluster centers, and scale per channel. We use these to quantize the data.
    Args:
        tensor_data: Tensor content as Numpy array.
        p: p-norm to use for the Lp-norm distance.
        n_bits: Number of bits to quantize the tensor.
        per_channel: Whether the quantization should be per-channel or not.
        channel_axis: Output channel index.
        n_iter: Number of iterations to search_methods for the optimal threshold.
        min_threshold: Minimal threshold to chose when the computed one is smaller.
        quant_error_method: an error function to optimize the parameters' selection accordingly (not used for this method).

    Returns:
        A dictionary containing the cluster assignments according to the k-means algorithm,
        the scales per channel and the multiplier num bits.
    """
    if n_bits > MULTIPLIER_N_BITS:
        raise Exception(f'Look-Up-Table bit configuration has {n_bits} bits. It must be less or equal to {MULTIPLIER_N_BITS}')
    # TODO: need to set this externally
    if len(np.unique(tensor_data.flatten())) < 2 ** n_bits:
        n_clusters = len(np.unique(tensor_data.flatten()))
    else:
        n_clusters = 2 ** n_bits
    kmeans = KMeans(n_clusters=n_clusters)
    axis_not_channel = [i for i in range(len(tensor_data.shape))]
    axis_not_channel.remove(channel_axis)
    if per_channel:
        scales_per_channel = np.max(np.abs(tensor_data), axis=tuple(axis_not_channel), keepdims=True)
    else:
        scales_per_channel = np.max(np.abs(tensor_data), keepdims=True)
    scales_per_channel = np.power(2.0, np.ceil(np.log2(scales_per_channel)))
    tensor_for_kmeans = int_quantization_with_scale(tensor_data, scales_per_channel, MULTIPLIER_N_BITS)
    kmeans.fit(tensor_for_kmeans.reshape(-1, 1))

    return {CLUSTER_CENTERS: np.round(kmeans.cluster_centers_),
            SCALE_PER_CHANNEL: scales_per_channel,
            }


# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tensorflow as tf

from model_compression_toolkit.common.hardware_representation import HardwareModel

if tf.__version__ < "2.6":
    from tensorflow.keras.layers import Conv2D, DepthwiseConv2D, Dense, Conv2DTranspose, Reshape, ZeroPadding2D, Dropout, \
        MaxPooling2D, Activation, ReLU, Add, PReLU, Flatten, Cropping2D
else:
    from keras.layers import Conv2D, DepthwiseConv2D, Dense, Conv2DTranspose, Reshape, ZeroPadding2D, \
    Dropout, MaxPooling2D, Activation, ReLU, Add, PReLU, Flatten, Cropping2D

from model_compression_toolkit.hardware_models.default_hwm import get_default_hardware_model

import model_compression_toolkit as mct
hwm = mct.hardware_representation


def get_default_hwm_keras():
    default_hwm = get_default_hardware_model()
    return generate_fhw_model_keras(name='default_hwm_keras',
                                    hardware_model=default_hwm)


def generate_fhw_model_keras(name: str, hardware_model: HardwareModel):
    """
    Generates a FrameworkHardwareModel object with default operation sets to layers mapping.
    Args:
        name: Name of the framework hardware model.
        hardware_model: HardwareModel object.
    Returns: a FrameworkHardwareModel object for the given HardwareModel.
    """

    fhwm_keras = hwm.FrameworkHardwareModel(hardware_model,
                                            name=name)
    with fhwm_keras:
        hwm.OperationsSetToLayers("NoQuantization", [Reshape,
                                                     tf.reshape,
                                                     Flatten,
                                                     Cropping2D,
                                                     ZeroPadding2D,
                                                     Dropout,
                                                     MaxPooling2D,
                                                     tf.split,
                                                     tf.quantization.fake_quant_with_min_max_vars])

        hwm.OperationsSetToLayers("Conv", [Conv2D,
                                           DepthwiseConv2D,
                                           tf.nn.conv2d,
                                           tf.nn.depthwise_conv2d])

        hwm.OperationsSetToLayers("FullyConnected", [Dense])

        hwm.OperationsSetToLayers("ConvTranspose", [Conv2DTranspose,
                                                    tf.nn.conv2d_transpose])

        hwm.OperationsSetToLayers("AnyReLU", [tf.nn.relu,
                                              tf.nn.relu6,
                                              hwm.LayerFilterParams(ReLU, negative_slope=0.0),
                                              hwm.LayerFilterParams(Activation, activation="relu")])

        hwm.OperationsSetToLayers("Add", [tf.add,
                                          Add])

        hwm.OperationsSetToLayers("PReLU", [PReLU])

        hwm.OperationsSetToLayers("Swish", [tf.nn.swish,
                                            hwm.LayerFilterParams(Activation, activation="swish")])

        hwm.OperationsSetToLayers("Sigmoid", [tf.nn.sigmoid,
                                              hwm.LayerFilterParams(Activation, activation="sigmoid")])

        hwm.OperationsSetToLayers("Tanh", [tf.nn.tanh,
                                           hwm.LayerFilterParams(Activation, activation="tanh")])
    return fhwm_keras



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
from typing import Callable, Any, List

from model_compression_toolkit import FrameworkInfo, MixedPrecisionQuantizationConfig
from model_compression_toolkit.common import Graph, BaseNode
from model_compression_toolkit.common.model_builder_mode import ModelBuilderMode


class SensitivityEvaluationManager:
    """
    Class to wrap and manage the computation on distance metric for Mixed-Precision quantization search.
    """

    def __init__(self,
                 graph: Graph,
                 fw_info: FrameworkInfo,
                 quant_config: MixedPrecisionQuantizationConfig,
                 representative_data_gen: Callable,
                 model_builder: Callable,
                 move_tensors_func: Callable = None):
        """
        Initiates all relevant objects to manage a sensitivity evaluation for MP search.
        Args:
            graph: Graph to search for its MP configuration.
            fw_info: FrameworkInfo object about the specific framework
                (e.g., attributes of different layers' weights to quantize).
            quant_config: MP Quantization configuration for how the graph should be quantized.
            representative_data_gen: Dataset used for getting batches for inference.
            model_builder: A function that builds a model object for the currently used framework.
            move_tensors_func: A function that moves tensors in list of tensors to cpu memory
                (relevant only when using Pytorch framework).
        """
        self.quant_config = quant_config

        # If used with Pytorch, we might need to move tensors to cpu before applying numpy functions
        self.move_tensors_func = move_tensors_func

        # Get interest points for distance measurement and a list of sorted configurable nodes names
        self.sorted_configurable_nodes_names = graph.get_configurable_sorted_nodes_names()
        self.interest_points = graph.get_configurable_sorted_nodes()

        # Build a mixed-precision model which can be configured to use different bitwidth in different layers.
        # And a baseline model.
        self.baseline_model, self.model_mp = build_models(graph, fw_info, self.interest_points, model_builder)

        # Build images batches for inference comparison
        self.images_batches = get_images_batches(quant_config.num_of_images, representative_data_gen)

        # Get baseline model inference on all samples
        self.baseline_tensors_list = []  # setting from outside scope

    def init_baseline_tensors_list(self):
        """
        Evaluates the baseline model on all images and saves the obtained lists of tensors in a list for later use.
        Initiates a class variable self.baseline_tensors_list
        """
        self._get_baseline_tensors()
        if self.move_tensors_func:
            # for Pytorch framework use only
            self.baseline_tensors_list = [self.move_tensors_func(tensors) for tensors in self.baseline_tensors_list]

    def _get_baseline_tensors(self):
        """
        Evaluate the baseline model on each of the images batches and constructs a list with the results for each batch.
        Sets a class variable with the results.
        """
        # Infer all images with baseline model, and save them as a list.
        # If the model contains only one output we save it a list. If it's a list already, we keep it as a list.
        self.baseline_tensors_list = [_tensors_as_list(self.baseline_model(images)) for images in self.images_batches]

    def compute_distance_matrix(self,
                                baseline_tensors: List[Any],
                                mp_tensors: List[Any]):
        """
        Compute the distance between the MP model's outputs and the baseline model's outputs
        for each image in the batch that was inferred.
        Args:
            baseline_tensors: Baseline model's output tensors.
            mp_tensors: MP model's output tensors.
        Returns:
            A distance matrix that maps each node's index to the distance between this node's output
             and the baseline model's output for all images that were inferred.
        """

        num_interest_points = len(baseline_tensors)
        num_samples = len(baseline_tensors[0])
        distance_matrix = np.ndarray((num_interest_points, num_samples))
        for i in range(num_interest_points):
            for j in range(num_samples):
                distance_matrix[i, j] = \
                    self.quant_config.compute_distance_fn(baseline_tensors[i][j].numpy(), mp_tensors[i][j].numpy())
        return distance_matrix

    def build_distance_metrix(self):
        """
        Builds a matrix that contains the distances between the baseline and MP models for each interest point.
        Returns: A distance matrix.
        """
        # List of distance matrices. We create a distance matrix for each sample from the representative_data_gen
        # and merge all of them eventually.
        distance_matrices = []

        # Compute the distance matrix for num_of_images images.
        for images, baseline_tensors in zip(self.images_batches, self.baseline_tensors_list):
            # when using model.predict(), it does not use the QuantizeWrapper functionality
            mp_tensors = _tensors_as_list(self.model_mp(images))

            if self.move_tensors_func:
                # in case the tensors are on GPU, and we use Pytorch,
                # need to move them to CPU in order to convert to numpy arrays
                mp_tensors = self.move_tensors_func(mp_tensors)

            # Build distance matrix: similarity between the baseline model to the float model
            # in every interest point for every image in the batch.
            distance_matrices.append(self.compute_distance_matrix(baseline_tensors, mp_tensors))

        # Merge all distance matrices into a single distance matrix.
        distance_matrix = np.concatenate(distance_matrices, axis=1)

        # Assert we used a correct number of images for computing the distance matrix
        assert distance_matrix.shape[1] == self.quant_config.num_of_images
        return distance_matrix


def build_models(graph: Graph, fw_info: FrameworkInfo, interest_points: List[BaseNode], model_builder: Callable) -> Any:
    """
    Args:
        graph: Graph to search for its MP configuration.
        fw_info: FrameworkInfo object about the specific framework
                (e.g., attributes of different layers' weights to quantize).
        interest_points: A sorted list of graph nodes that constitute as points to compute distance measurement.
        model_builder:
    Returns: A tuple with two models built from the given graph: a baseline model (with baseline configuration) and
        an MP model (which can be configured for a specific bitwidth configuration).
        Note that the type of the returned models is dependent on the used framework (TF/Pytorch).
    """
    # Build a mixed-precision model which can be configured to use different bitwidth in different layers.
    model_mp, _ = model_builder(graph,
                                mode=ModelBuilderMode.MIXEDPRECISION,
                                append2output=interest_points,
                                fw_info=fw_info)

    # Build a baseline model.
    baseline_model, _ = model_builder(graph,
                                      mode=ModelBuilderMode.FLOAT,
                                      append2output=interest_points)

    return baseline_model, model_mp,


def get_images_batches(num_of_images: int, representative_data_gen: Callable) -> List[Any]:
    """
    Construct batches of image samples for inference.
    Args:
        num_of_images: Num of total images for evaluation.
        representative_data_gen: ataset used for getting batches for inference.
    Returns: A list of images batches (lists of images)
    """
    # First, select images to use for all measurements.
    samples_count = 0  # Number of images we used so far to compute the distance matrix.
    images_batches = []
    while samples_count < num_of_images:
        # Get a batch of images to infer in both models.
        inference_batch_input = representative_data_gen()
        batch_size = inference_batch_input[0].shape[0]

        # If we sampled more images than we should use in the distance matrix,
        # we take only a subset of these images and use only them for computing the distance matrix.
        if batch_size > num_of_images - samples_count:
            inference_batch_input = [x[:num_of_images - samples_count] for x in inference_batch_input]
            assert num_of_images - samples_count == inference_batch_input[0].shape[0]
            batch_size = num_of_images - samples_count

        images_batches.append(inference_batch_input)
        samples_count += batch_size
    return images_batches


def _tensors_as_list(tensors: Any) -> List[Any]:
    """
    Create a list of tensors if they are not in a list already.
    This functions helpful when the graph has only one node (so the model's output is a Tensor and not a list of
    Tensors).
    Args:
        tensors: Tensors to return as a list.
    Returns:
        List of tensors.
    """
    if not isinstance(tensors, list):
        return [tensors]
    return tensors


def compute_mp_distance_measure(distance_matrix, metrics_weights_fn):
    """
    Computes the final distance value out of a distance matrix.
    Args:
        distance_matrix: A matrix that contains the distances between the baseline and MP models
            for each interest point.
        metrics_weights_fn:
    Returns: Distance value.
    """
    # Compute the distance between the baseline model's outputs and the MP model's outputs.
    # The distance is the mean of distances over all images in the batch that was inferred.
    mean_distance_per_layer = distance_matrix.mean(axis=1)
    # Use weights such that every layer's distance is weighted differently (possibly).
    return np.average(mean_distance_per_layer, weights=metrics_weights_fn(distance_matrix))


# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import operator

import torch
from torch import add, flatten, reshape, split, unsqueeze, dropout, sigmoid, tanh
from torch.nn import Conv2d, ConvTranspose2d, Linear, BatchNorm2d
from torch.nn import Dropout, Flatten, Hardtanh
from torch.nn import ReLU, ReLU6, PReLU, SiLU, Sigmoid, Tanh
from torch.nn.functional import relu, relu6, prelu, silu, hardtanh

from model_compression_toolkit.common.hardware_representation import HardwareModel
from model_compression_toolkit.common.hardware_representation.hardware2framework import \
    FrameworkHardwareModel, LayerFilterParams
from model_compression_toolkit.common.hardware_representation.hardware2framework import \
    OperationsSetToLayers
from model_compression_toolkit.hardware_models.default_hwm import get_default_hardware_model


def get_default_hwm_pytorch():
    default_hwm = get_default_hardware_model()
    return generate_fhw_model_pytorch(name='default_hwm_pytorch',
                                      hardware_model=default_hwm)


def generate_fhw_model_pytorch(name: str, hardware_model: HardwareModel):
    """
    Generates a FrameworkHardwareModel object with default operation sets to layers mapping.
    Args:
        name: Name of the framework hardware model.
        hardware_model: HardwareModel object.
    Returns: a FrameworkHardwareModel object for the given HardwareModel.
    """

    fhwm_pytorch = FrameworkHardwareModel(hardware_model,
                                          name=name)

    with fhwm_pytorch:
        OperationsSetToLayers("NoQuantization", [Dropout,
                                                 Flatten,
                                                 dropout,
                                                 flatten,
                                                 split,
                                                 operator.getitem,
                                                 reshape,
                                                 unsqueeze,
                                                 BatchNorm2d,
                                                 torch.Tensor.size])

        OperationsSetToLayers("Conv", [Conv2d])

        OperationsSetToLayers("FullyConnected", [Linear])

        OperationsSetToLayers("ConvTranspose", [ConvTranspose2d])

        OperationsSetToLayers("AnyReLU", [torch.relu,
                                          ReLU,
                                          ReLU6,
                                          relu,
                                          relu6,
                                          LayerFilterParams(Hardtanh, min_val=0),
                                          LayerFilterParams(hardtanh, min_val=0)])

        OperationsSetToLayers("Add", [operator.add,
                                      add])

        OperationsSetToLayers("PReLU", [PReLU,
                                        prelu])

        OperationsSetToLayers("Swish", [SiLU,
                                        silu])

        OperationsSetToLayers("Sigmoid", [Sigmoid,
                                          sigmoid])

        OperationsSetToLayers("Tanh", [Tanh,
                                       tanh])

    return fhwm_pytorch



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from model_compression_toolkit.common.gptq.gptq_config import GradientPTQConfig
from model_compression_toolkit.common.quantization import quantization_config
from model_compression_toolkit.common.mixed_precision import mixed_precision_quantization_config
from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig, \
    QuantizationErrorMethod, DEFAULTCONFIG
from model_compression_toolkit.common import hardware_representation
from model_compression_toolkit.hardware_models.get_model import get_model
from model_compression_toolkit.common.mixed_precision.kpi import KPI
from model_compression_toolkit.common.mixed_precision.mixed_precision_quantization_config import MixedPrecisionQuantizationConfig
from model_compression_toolkit.common.logger import set_log_folder
from model_compression_toolkit.common.data_loader import FolderImageLoader
from model_compression_toolkit.common.framework_info import FrameworkInfo, ChannelAxis
from model_compression_toolkit.common.defaultdict import DefaultDict
from model_compression_toolkit.common import network_editors as network_editor

from model_compression_toolkit.keras.quantization_facade import keras_post_training_quantization, \
    keras_post_training_quantization_mixed_precision
from model_compression_toolkit.keras.quantization_facade import get_keras_gptq_config
from model_compression_toolkit.pytorch.quantization_facade import pytorch_post_training_quantization, \
    pytorch_post_training_quantization_mixed_precision

__version__ = "1.3.0"



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np

import model_compression_toolkit.common.quantization.quantization_config as qc
from model_compression_toolkit.common.constants import MIN_THRESHOLD, THRESHOLD
from model_compression_toolkit.common.quantization.quantization_params_generation.qparams_search import \
    qparams_selection_tensor_search, qparams_selection_histogram_search
from model_compression_toolkit.common.similarity_analyzer import compute_mse


def mse_selection_tensor(tensor_data: np.ndarray,
                         p: int,
                         n_bits: int,
                         per_channel: bool = False,
                         channel_axis: int = 1,
                         n_iter: int = 10,
                         min_threshold: float = MIN_THRESHOLD,
                         quant_error_method: qc.QuantizationErrorMethod = qc.QuantizationErrorMethod.MSE) -> dict:
    """
    Compute the optimal threshold based on mean square error (MSE) to quantize the tensor.
    The threshold is constrained and the search is done iteratively over n_iter thresholds.
    A minimal threshold to choose when the computed threshold is smaller can be passed.

    Args:
        tensor_data: Tensor content as Numpy array.
        p: p-norm to use for the Lp-norm distance (not used for this method).
        n_bits: Number of bits to quantize the tensor.
        per_channel: Whether the quantization should be per-channel or not.
        channel_axis: Output channel index.
        n_iter: Number of iterations to search for the optimal threshold.
        min_threshold: Minimal threshold to chose when the computed one is smaller.
        quant_error_method: an error function to optimize the parameters' selection accordingly (not used for this method).

    Returns:
        Optimal threshold to quantize the tensor based on MSE.
    """

    return {THRESHOLD: qparams_selection_tensor_search(lambda x, y, threshold: compute_mse(x, y),
                                                       tensor_data,
                                                       n_bits,
                                                       per_channel=per_channel,
                                                       channel_axis=channel_axis,
                                                       n_iter=n_iter,
                                                       min_threshold=min_threshold)}


def mse_selection_histogram(bins: np.ndarray,
                            counts: np.ndarray,
                            p: int,
                            n_bits: int,
                            min_value: float,
                            max_value: float,
                            constrained: bool = True,
                            n_iter: int = 10,
                            min_threshold: float = MIN_THRESHOLD,
                            quant_error_method: qc.QuantizationErrorMethod = qc.QuantizationErrorMethod.MSE) -> dict:
    """
    Compute the optimal threshold based on the mean square error (MAE) to quantize a histogram.
    The threshold is either constrained or unconstrained and the search is done iteratively over
    n_iter thresholds. Minimal threshold to use if threshold is too small can be passed.

    Args:
        bins: Bins values of the histogram.
        counts: Bins counts of the histogram.
        p: p-norm to use for the Lp-norm distance (not used for this method).
        n_bits: Number of bits to quantize the tensor.
        min_value: Min value (not used for this method).
        max_value: Max value (not used for this method).
        constrained: Whether the threshold should be constrained or not.
        n_iter: Number of iteration ot search for the threshold.
        min_threshold: Minimal threshold to use if threshold is too small.
        quant_error_method: an error function to optimize the parameters' selection accordingly (not used for this method).

    Returns:
        Optimal threshold to quantize the histogram based on the MSE.
    """
    return {THRESHOLD: qparams_selection_histogram_search(lambda q_bins, q_count, _bins, _counts, threshold, _range:
                                                          _mse_error_histogram(q_bins, q_count, _bins, _counts),
                                                          bins,
                                                          counts,
                                                          n_bits,
                                                          constrained=constrained,
                                                          n_iter=n_iter,
                                                          min_threshold=min_threshold)}


def _mse_error_histogram(q_bins: np.ndarray,
                         q_count: np.ndarray,
                         bins: np.ndarray,
                         counts: np.ndarray) -> np.float:
    """
    Compute the error function between a histogram to its quantized version.
    The error is computed based on the mean square error the distributions have.

    Args:
        q_bins: Bins values of the quantized histogram.
        q_count: Bins counts of the quantized histogram.
        bins: Bins values of the original histogram.
        counts: Bins counts of the original histogram.

    Returns:
        MSE between the two histograms.
    """

    return np.sum((np.power((q_bins - bins)[:-1], 2.0) * counts)) / np.sum(counts)
